/*
 * Copyright 2000-2001 Hans Reiser
 */

#include "fsck.h"


/* fixme: what should be done when not all transactions can be replayed in
   proper order? */
int reiserfs_replay_journal (reiserfs_filsys_t * fs)
{
    struct buffer_head * bh;
    struct reiserfs_journal_header * j_head;
    reiserfs_trans_t cur, newest, control;


    bh = bread (fs->s_journal_dev, get_sb_journal_1st_block (fs->s_ondisk_sb) +
		get_sb_journal_size (fs->s_ondisk_sb),
		fs->s_blocksize);
    if (!bh) {
	fsck_progress ("Could not read journal header\n");
	return 0;
    }

    j_head = (struct reiserfs_journal_header *)(bh->b_data);
    control.mount_id = get_jh_mount_id (j_head);
    control.trans_id = get_jh_last_flushed (j_head);
    control.desc_blocknr = get_jh_replay_start_offset (j_head);

    if (!get_boundary_transactions (fs, &cur, &newest)) {
	fsck_progress ("No transactions found\n");
	brelse (bh);
	return 0;
    }

    if (control.mount_id < newest.mount_id) {
	fsck_progress ("Journal header's mountid seems wrong (%lu), "
		       "replaying all transactions with mountid %lu\n",
		       control.mount_id, newest.mount_id);
	control.mount_id = newest.mount_id;
	control.trans_id = 0;
    }

    while (1) {
	if (cur.mount_id == control.mount_id &&
	    cur.trans_id > control.trans_id) {
	    /* transaction is of */
	    replay_one_transaction (fs, &cur, fsck_log_file (fs));
	    set_jh_last_flushed (j_head, cur.trans_id);
	    mark_buffer_dirty (bh);
	}
	if (!next_transaction (fs, &cur))
	    break;
    }

    brelse (bh);
    return 0;
}




#if 0
int reiserfs_replay_journal2 (reiserfs_filsys_t * fs)
{
    struct buffer_head * d_bh, * c_bh, * jh_bh;
    struct reiserfs_journal_header * j_head;
    struct reiserfs_journal_desc * j_desc;
    struct reiserfs_journal_commit * j_commit;
    unsigned long latest_mount_id;
    unsigned long j_cur;
    unsigned long j_start;
    unsigned long j_size;
    unsigned long mount_id, trans_id;
    unsigned long t_first, t_last, t_count, t_flushed;
    unsigned long t_offset;
    int i;

    fsck_progress ("Analyzing journal..");

    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);
    j_cur = 0;
    j_size = get_sb_journal_size (fs->s_ondisk_sb);
    t_first = 0;
    t_last = 0;
    latest_mount_id = 0;

    /* look for the transactions with the most recent mount_id */
    for (j_cur = 0; j_cur < j_size; ) {
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	if (is_valid_transaction (fs, d_bh))
	    /*if (d_bh && who_is_this (d_bh->b_data, d_bh->b_size) == THE_JDESC &&
	      is_valid_transaction (fs, d_bh))*/ {
	    j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;

	    mount_id = get_jdesc_mount_id (j_desc);
	    trans_id = get_jdesc_trans_id (j_desc);

	    if (mount_id > latest_mount_id) {
		/* more recent mount_id found */
		latest_mount_id = mount_id;
		t_first = t_last = trans_id;
		t_offset = j_cur;
		t_count = 1;
	    } else if (mount_id == latest_mount_id) {
		t_count ++;
		if (trans_id > t_last)
		    t_last = trans_id;
		if (trans_id < t_first) {
		    t_first = trans_id;
		    t_offset = j_cur;
		}
	    }
	    j_cur += get_jdesc_len (j_desc) + 1;
	}
	j_cur ++;
	brelse (d_bh);
    }

    /* replay only if journal header looks resonable */
    jh_bh = bread (fs->s_journal_dev, j_start + j_size, fs->s_blocksize);
    j_head = (struct reiserfs_journal_header *)(jh_bh->b_data);

    if (latest_mount_id != le32_to_cpu (j_head->j_mount_id)) {
	fsck_progress ("nothing to replay (no transactions match to latest mount id)\n");
	brelse (jh_bh);
	return 0;
    }
    /* last transaction flushed - which should not be replayed */
    t_flushed = le32_to_cpu (j_head->j_last_flush_trans_id);
    if (t_flushed >= t_last) {
	fsck_progress ("nothing to replay (no transactions older than last flushed one found)\n");
	brelse (jh_bh);
	return 0;
    }
    if (t_first > t_flushed + 1) {
	if (t_flushed)
	    fsck_progress ("last flushed trans %lu, the oldest but newer is %lu\n",
			   t_flushed, t_first);
    } else {
	/* start replaying with first not flushed transaction */
	t_first = t_flushed + 1;
	t_offset = le32_to_cpu (j_head->j_first_unflushed_offset);
    }

    fsck_progress ("last flushed trans %lu, mount_id %lu, "
		   "will replay from %lu up to %lu:Yes?",
		   t_flushed, latest_mount_id, t_first, t_last);
    if (!fsck_user_confirmed (fs, "", "Yes\n", 1))
	die ("");


    

    /* replay transactions we have found */
    for (j_cur = t_offset; t_first <= t_last; t_first ++) {
	unsigned long offset;
	
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	replay_one_transaction (fsck_log (fs), fs, d_bh);
	brelse (d_bh);
	brelse (c_bh);
	j_cur ++;
	j_cur %= j_size;

	/* update journal header */
	j_head->j_last_flush_trans_id = cpu_to_le32 (t_first);
	mark_buffer_dirty (jh_bh);
	bwrite (jh_bh);
    }

	j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;
	if (who_is_this (d_bh->b_data, d_bh->b_size) != THE_JDESC ||
	    get_jdesc_mount_id (j_desc) != latest_mount_id ||
	    get_jdesc_trans_id (j_desc) != t_first)
	    die ("reiserfs_replay_journal: desc block not found");

	offset = j_cur + 1;
	j_cur += get_jdesc_len (j_desc) + 1;
	j_cur %= j_size;
	c_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	j_commit = (struct reiserfs_journal_commit *)c_bh->b_data;
	if (!does_desc_match_commit (j_desc, j_commit)) {
	    print_block (stderr, fs, d_bh);
	    reiserfs_warning (stderr, "%d %d\n", get_jcommit_len (j_commit),
			      get_jcommit_trans_id (j_commit));
	    die ("reiserfs_replay_journal: commit block not found");
	}

	fsck_log ("Mount_id %lu, transaction %lu, desc block %lu, commit block %lu: (",
		  latest_mount_id, t_first, d_bh->b_blocknr, c_bh->b_blocknr);

	replay_one_transaction ();
	/* replay one transaction */
	for (i = 0; i < get_jdesc_len (j_desc); i ++) {
	    struct buffer_head * in_place, * log;
	    unsigned long block;

	    log = bread (fs->s_journal_dev, j_start + ((offset + i) % j_size), fs->s_blocksize);

	    if (i < JOURNAL_TRANS_HALF) {
		block = le32_to_cpu (j_desc->j2_realblock[i]);
	    } else {
		block = le32_to_cpu (j_commit->j3_realblock[i - JOURNAL_TRANS_HALF]);
	    }

	    if (not_journalable (fs, block)) {
		fsck_log ("transaction %lu, block %d could not be replayed (%lu)\n",
			      t_first, i, block);
	    } else {
		fsck_log (" %lu", block);
		
		in_place = getblk (fs->s_dev, block, fs->s_blocksize) ;
		memcpy (in_place->b_data, log->b_data, fs->s_blocksize);
		mark_buffer_dirty (in_place);
		mark_buffer_uptodate (in_place, 1);
		bwrite (in_place);
		brelse (in_place);
	    }
	    brelse (log);
	    
	}
	fsck_log (")\n");

	brelse (d_bh);
	brelse (c_bh);
	j_cur ++;
	j_cur %= j_size;

	/* update journal header */
	j_head->j_last_flush_trans_id = cpu_to_le32 (t_first);
	mark_buffer_dirty (jh_bh);
	bwrite (jh_bh);
    }

    brelse (jh_bh);
    fsck_progress ("Journal replaied\n");
    return 0;
}
#endif
